/*div.documentwrapper{
    background-color: #00617E;
}

div.body{
    background-color: unset;
}

div.document{
    margin-top: 40px;
}
*/

body{
    background: linear-gradient(#0D4960,#00617E);
    font-family: 'Overpass', sans-serif;
    color: white;
    margin-left: auto;
    margin-right: auto;
    max-width: 800px;
    padding-left: 15px;
    padding-right: 15px;
}


.logo{
    margin-left: auto;
    margin-right: auto;
    max-width: 700px;
    max-width: 700px;
    display: block;
}

h1{
    text-align: center;
    max-width: 800px;
    margin-top: 0px;
    margin-left: auto;
    margin-right: auto;
    margin-bottom: 50px;
}

h2{
    text-align: center;
    font-size: 55px;
    margin-bottom: 25px;
    margin-top: 30px;
    font-weight: 400;
}

h3{
    font-size: 40px;
    margin-bottom: 15px;
    margin-right: 0px;
}

a{
    color: white;
    text-decoration: none;
    cursor: pointer;
}


p{
    font-size: 25px;
    line-height: 30px;
}

ul{
    font-size: 25px;
    line-height: 30px;
}

br{
    height: 10px;
}

.big-button{
    background: #0D4960;
    border: 2px;
    border-color: white;
    border-style: solid;
    padding-left: 25px;
    padding-right: 20px;
    padding-bottom: 20px;
    padding-top: 5px;
    margin-bottom: 20px;
    display: block;
}

a:hover .big-button {
    background-color: #00617E;
}

.text-link{
    text-decoration: underline;
   
}

.small-link{
    background: #0D4960;
    font-size: 25px;
    margin-bottom: 20px;
    border: 2px;
    border-color: white;
    border-style: solid;
    padding: 15px;
    max-width: 800;
    text-align: center;
    font-weight: bold;
    display: block;
}


.small-link:hover{
    background-color: #00617E;
}

.big-button ORANGE{
    background: #C75109;
    cursor: pointer;
    padding-left: 20px;
    padding-right: 20px;
    padding-bottom: 20px;
    padding-top: 5px;
    margin-bottom: 20px;
}

.footer{
    margin-top: 100px;
    margin-bottom: 30px;
}
