

div.wy-side-nav-search{
    background: #00617E;
}

div.wy-side-scroll{
    background: #0D4960;
}

a.reference:hover{
    background: rgb(11, 61, 78);
}

.logo{
    width:120px !important;
    margin-left:auto !important;
    margin-right: auto !important;
    padding-bottom: 20px !important;
    margin-top: 15px !important;
}

.intro{
    color: aqua;
}

.heading{
    color: #0D4960;
    font-size: 25px;
    font-weight: 800;
    line-height: 30px;

}

